# Copyright (C) 2019 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from adi.rx_tx import rx_tx
from adi.context_manager import context_manager


class ad9371(rx_tx, context_manager):
    """ AD9371 Transceiver """

    complex_data = True
    rx_channel_names = ["voltage0_i", "voltage0_q", "voltage1_i", "voltage1_q"]
    tx_channel_names = ["voltage0", "voltage1", "voltage2", "voltage3"]
    device_name = ""
    rx_enabled_channels = [0, 1]
    tx_enabled_channels = [0, 1]

    def __init__(self, uri=""):

        context_manager.__init__(self, uri, self.device_name)

        self.ctrl = self.ctx.find_device("ad9371-phy")
        self.rxadc = self.ctx.find_device("axi-ad9371-rx-hpc")
        self.rxobs = self.ctx.find_device("axi-ad9371-rx-obs-hpc")
        self.txdac = self.ctx.find_device("axi-ad9371-tx-hpc")

        rx_tx.__init__(self, self.rx_enabled_channels, self.tx_enabled_channels)

    @property
    def gain_control_mode(self):
        """gain_control_mode: Mode of receive path AGC. Options are:
        automatic, hybrid, manual"""
        return self.get_iio_attr("voltage0", "gain_control_mode", False)

    @gain_control_mode.setter
    def gain_control_mode(self, value):
        self.set_iio_attr_str("voltage0", "gain_control_mode", False, value)

    @property
    def rx_hardwaregain(self):
        """rx_hardwaregain: Gain applied to RX path. Only applicable when
        gain_control_mode is set to 'manual'"""
        return self.get_iio_attr("voltage0", "hardwaregain", False)

    @rx_hardwaregain.setter
    def rx_hardwaregain(self, value):
        if self.gain_control_mode == "manual":
            self.set_iio_attr("voltage0", "hardwaregain", False, value)

    @property
    def tx_hardwaregain(self):
        """tx_hardwaregain: Attenuation applied to TX path"""
        return self.get_iio_attr("voltage0", "hardwaregain", True)

    @tx_hardwaregain.setter
    def tx_hardwaregain(self, value):
        self.set_iio_attr("voltage0", "hardwaregain", True, value)

    @property
    def rx_rf_bandwidth(self):
        """rx_rf_bandwidth: Bandwidth of front-end analog filter of RX path"""
        return self.get_iio_attr("voltage0", "rf_bandwidth", False)

    @property
    def tx_rf_bandwidth(self):
        """tx_rf_bandwidth: Bandwidth of front-end analog filter of TX path"""
        return self.get_iio_attr("voltage0", "rf_bandwidth", True)

    @property
    def rx_sample_rate(self):
        """rx_sample_rate: Sample rate RX path in samples per second"""
        return self.get_iio_attr("voltage0", "sampling_frequency", False)

    @property
    def tx_sample_rate(self):
        """tx_sample_rate: Sample rate TX path in samples per second"""
        return self.get_iio_attr("voltage0", "sampling_frequency", True)

    @property
    def rx_lo(self):
        """rx_lo: Carrier frequency of RX path"""
        return self.get_iio_attr("altvoltage0", "RX_LO_frequency", True)

    @rx_lo.setter
    def rx_lo(self, value):
        self.set_iio_attr("altvoltage0", "RX_LO_frequency", True, value)

    @property
    def tx_lo(self):
        """tx_lo: Carrier frequency of TX path"""
        return self.get_iio_attr("altvoltage1", "TX_LO_frequency", True)

    @tx_lo.setter
    def tx_lo(self, value):
        self.set_iio_attr("altvoltage1", "TX_LO_frequency", True, value)

    @property
    def sn_lo(self):
        """sn_lo: Carrier frequency of Sniffer/ORx path"""
        return self.get_iio_attr("altvoltage2", "RX_SN_LO_frequency", True)

    @sn_lo.setter
    def sn_lo(self, value):
        self.set_iio_attr("altvoltage2", "RX_SN_LO_frequency", True, value)



# Copyright (C) 2019 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
from adi.dds import dds
from adi.attribute import attribute
import iio


class phy(attribute):
    ctrl = []

    def __del__(self):
        self.ctrl = []


class rx(attribute):
    """ Buffer handling for receive devices """

    rxadc = []
    rx_channel_names = []
    rxbuf = None

    def __init__(self, rx_enabled_channels, rx_buffer_size=1024):
        self.rx_enabled_channels = rx_enabled_channels
        self.rx_buffer_size = rx_buffer_size

    @property
    def rx_buffer_size(self):
        """rx_buffer_size: Size of receive buffer in samples"""
        return self.__rx_buffer_size

    @rx_buffer_size.setter
    def rx_buffer_size(self, value):
        self.__rx_buffer_size = value

    @property
    def rx_enabled_channels(self):
        """rx_enabled_channels: List of enabled channels (channel 1 is 0)"""
        return self.__rx_enabled_channels

    @rx_enabled_channels.setter
    def rx_enabled_channels(self, value):
        if self.complex_data:
            if max(value) > ((self.num_rx_channels) / 2 - 1):
                raise Exception("RX mapping exceeds available channels")
        else:
            if max(value) > ((self.num_rx_channels) - 1):
                raise Exception("RX mapping exceeds available channels")
        self.__rx_enabled_channels = value

    @property
    def num_rx_channels_enabled(self):
        return len(self.__rx_enabled_channels)

    def __del__(self):
        self.rxbuf = []
        self.rxadc = []

    def init_channels(self):
        if self.complex_data:
            for map in self.rx_enabled_channels:
                v = self.rxadc.find_channel(self.rx_channel_names[map * 2])
                v.enabled = True
                v = self.rxadc.find_channel(self.rx_channel_names[map * 2 + 1])
                v.enabled = True
        else:
            for map in self.rx_enabled_channels:
                v = self.rxadc.find_channel(self.rx_channel_names[map])
                v.enabled = True
        self.rxbuf = iio.Buffer(self.rxadc, self.__rx_buffer_size, False)

    def rx_complex(self):
        if not self.rxbuf:
            self.init_channels(False)
        self.rxbuf.refill()
        data = self.rxbuf.read()
        x = np.frombuffer(data, dtype=np.int16)
        indx = 0
        sig = []
        l = len(self.rx_enabled_channels) * 2
        for c in range(l // 2):
            sig.append(x[indx::l] + 1j * x[indx + 1 :: l])
            indx = indx + 2
        # Don't return list if a single channel
        if indx == 2:
            return sig[0]
        return sig

    def rx_non_complex(self):
        if not self.rxbuf:
            self.init_channels(False)
        self.rxbuf.refill()
        data = self.rxbuf.read()
        x = np.frombuffer(data, dtype=np.int16)
        indx = 0
        sig = []
        l = len(self.__rx_enabled_channels)
        for c in range(l):
            sig.append(x[c::l])
        # Don't return list if a single channel
        if self.__rx_enabled_channels == 1:
            return sig[0]
        return sig

    def rx(self):
        if self.complex_data:
            return self.rx_complex()
        else:
            return self.rx_non_complex()


class tx(dds, attribute):
    """ Buffer handling for receive devices """

    txdac = []
    tx_channel_names = []
    tx_buffer_size = 1024
    txbuf = None

    def __init__(self, tx_enabled_channels, tx_cyclic_buffer=False):
        self.tx_enabled_channels = tx_enabled_channels
        self.tx_cyclic_buffer = tx_cyclic_buffer
        dds.__init__(self)

    def __del__(self):
        self.txdac = []

    @property
    def tx_cyclic_buffer(self):
        """tx_cyclic_buffer: Enable cyclic buffer for TX"""
        return self.__tx_cyclic_buffer

    @tx_cyclic_buffer.setter
    def tx_cyclic_buffer(self, value):
        self.__tx_cyclic_buffer = value

    @property
    def num_tx_channels_enabled(self):
        return len(self.tx_enabled_channels)

    @property
    def tx_enabled_channels(self):
        """tx_enabled_channels: List of enabled channels (channel 1 is 0)"""
        return self.__tx_enabled_channels

    @tx_enabled_channels.setter
    def tx_enabled_channels(self, value):
        if self.complex_data:
            if max(value) > ((self.num_tx_channels) / 2 - 1):
                raise Exception("TX mapping exceeds available channels")
        else:
            if max(value) > ((self.num_tx_channels) - 1):
                raise Exception("TX mapping exceeds available channels")
        self.__tx_enabled_channels = value

    def init_channels(self):
        if self.complex_data:
            for map in self.tx_enabled_channels:
                v = self.txdac.find_channel(self.tx_channel_names[map * 2], True)
                v.enabled = True
                v = self.txdac.find_channel(self.tx_channel_names[map * 2 + 1], True)
                v.enabled = True
        else:
            for map in self.tx_enabled_channels:
                v = self.txdac.find_channel(self.tx_channel_names[map])
                v.enabled = True
        self.txbuf = iio.Buffer(
            self.txdac, self.tx_buffer_size, self.__tx_cyclic_buffer
        )

    def tx(self, data_np):
        if self.complex_data:
            if self.num_tx_channels_enabled == 1:
                data_np = [data_np]

            if len(data_np) != self.num_tx_channels_enabled:
                raise Exception("Not enough data provided for channel mapping")

            indx = 0
            l = self.num_tx_channels_enabled * 2
            data = np.empty(l * len(data_np[0]), dtype=np.int16)
            for chan in data_np:
                i = np.real(chan)
                q = np.imag(chan)
                data[indx::l] = i.astype(int)
                data[indx + 1 :: l] = q.astype(int)
                indx = indx + 2
        else:
            if self.num_tx_channels_enabled == 1:
                data_np = [data_np]
            indx = 0
            l = self.num_tx_channels_enabled
            iq = np.empty(l * len(data_np), dtype=np.int16)
            for chan in data_np:
                data[indx::l] = chan.astype(int)
                indx = indx + 1

        if not self.txbuf:
            self.disable_dds()
            self.tx_buff_length = len(data)
            self.init_channels(True)

        if len(data) != self.tx_buff_length:
            raise

        # Send data to buffer
        self.txbuf.write(bytearray(data))
        self.txbuf.push()


class rx_tx(rx, tx, phy):

    complex_data = False

    def __init__(self, rx_enabled_channels, tx_enabled_channels):
        self.num_rx_channels = len(self.rx_channel_names)
        self.num_tx_channels = len(self.tx_channel_names)
        rx.__init__(self, rx_enabled_channels)
        tx.__init__(self, tx_enabled_channels)

    def __del__(self):
        rx.__del__(self)
        tx.__del__(self)
        phy.__del__(self)

    def init_channels(self, istx=False):
        if istx:
            tx.init_channels(self)
        else:
            rx.init_channels(self)



# Copyright (C) 2019 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import print_function

import logging

import unittest
import numpy as np
from adi import Pluto
import iio


class TestPluto(unittest.TestCase):

    do_plots = False

    def freq_est(self, y, fs):
        N = len(y)
        T = 1.0 / fs
        yf = np.fft.fft(y)
        yf = np.fft.fftshift(yf)
        xf = np.linspace(-1.0 / (2.0 * T), 1.0 / (2.0 * T), N)
        if self.do_plots:
            import matplotlib.pyplot as plt

            fig, ax = plt.subplots()
            ax.plot(xf, 2.0 / N * np.abs(yf))
            plt.show()
        indx = np.argmax(np.abs(yf))
        return xf[indx]

    def setUp(self):
        self.longMessage = True

    def tearDown(self):
        pass

    def testPlutoADC(self):
        # See if we can get non-zero data from Pluto
        sdr = Pluto()
        data = sdr.rx()
        s = np.sum(np.abs(data))
        self.assertGreater(s, 0, "check non-zero data")

    def testPlutoDAC(self):
        # See if we can tone from Pluto using DMAs
        sdr = Pluto()
        sdr.tx_lo = 1000000000
        sdr.rx_lo = 1000000000
        sdr.tx_cyclic_buffer = True
        sdr.tx_hardwaregain = -30
        sdr.gain_control_mode = "slow_attack"
        sdr.rx_buffer_size = 2 ** 20
        # Create a sinewave waveform
        RXFS = int(sdr.sample_rate)
        fc = RXFS * 0.1
        N = 2 ** 15
        ts = 1 / float(RXFS)
        t = np.arange(0, N * ts, ts)
        i = np.cos(2 * np.pi * t * fc) * 2 ** 15 * 0.5
        q = np.sin(2 * np.pi * t * fc) * 2 ** 15 * 0.5
        iq = i + 1j * q
        # Pass through SDR
        sdr.tx(iq)
        for k in range(5):
            data = sdr.rx()

        tone_freq = self.freq_est(data, RXFS)

        if self.do_plots:
            import matplotlib.pyplot as plt

            reals = np.real(data)
            plt.plot(reals)
            imags = np.imag(data)
            plt.plot(imags)
            plt.xlabel("Samples")
            plt.ylabel("Amplitude [dbFS]")
            plt.show()

        diff = np.abs(tone_freq - fc)
        self.assertGreater(fc * 0.01, diff, "Frequency offset")


if __name__ == "__main__":
    from os import path
    import sys

    try:
        iio.Context("ip:192.168.2.1")
    except:
        print("No Pluto found")
        sys.exit(1)
    unittest.main()



# Copyright (C) 2019 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from adi.attribute import attribute
import numpy as np


class dds(attribute):
    """ DDS Signal generators: Each reference design contains two DDSs per channel.
        this allows for two complex tones to be generated per complex channel.
    """

    dds_frequencies = []
    dds_scales = []
    dds_phases = []
    dds_enabled = []

    def __init__(self):
        self.dds_frequencies = np.zeros(self.num_tx_channels * 2)
        self.dds_scales = np.zeros(self.num_tx_channels * 2)
        self.dds_phases = np.zeros(self.num_tx_channels * 2)
        self.dds_enabled = np.zeros(self.num_tx_channels * 2, dtype=bool)

    def update_dds(self, attr, value):
        indx = 0
        for chan in self.txdac.channels:
            if not chan.name:
                continue
            if attr == "raw":
                chan.attrs[attr].value = str(int(value[indx]))
            else:
                chan.attrs[attr].value = str(value[indx])
            indx = indx + 1

    def read_dds(self, attr):
        values = []
        indx = 0
        for chan in self.txdac.channels:
            if not chan.name:
                continue
            values.append(chan.attrs[attr].value)
            indx = indx + 1
        return values

    def disable_dds(self):
        self.dds_enabled = np.zeros(self.num_tx_channels * 2, dtype=bool)

    @property
    def dds_frequencies(self):
        """ Frequencies of DDSs in Hz"""
        return self.read_dds("frequency")

    @dds_frequencies.setter
    def dds_frequencies(self, value):
        self.update_dds("frequency", value)

    @property
    def dds_scales(self):
        """ Scale of DDS signal generators
            Ranges [0,1]
        """
        return self.read_dds("scale")

    @dds_scales.setter
    def dds_scales(self, value):
        self.update_dds("scale", value)

    @property
    def dds_phases(self):
        """ Phases of DDS signal generators.
            Range in millidegrees [0,360000]
        """
        return self.read_dds("phase")

    @dds_scales.setter
    def dds_phases(self, value):
        self.update_dds("phase", value)

    @property
    def dds_enabled(self):
        """ DDS generator enable state """
        return self.read_dds("raw")

    @dds_enabled.setter
    def dds_enabled(self, value):
        self.update_dds("raw", value)
